/* jumpstart.c -- run a process with shared libraries already mapped.
   Copyright 2007 BitWagon Software LLC.  All rights reserved.
   Licensed under GNU General Public License, version 2 [GPLv2].
*/
// This program "plays around under the covers" with things that are not
// supported as external interfaces.  Some aspects depend on un-intended
// "features" of glibc versions 2.3.6 through 2.6.1.
// This implementation is specific to x86 architecture.
//
// The whole point of jumpstart is runtime efficiency.  Jumpstart puts
// a premium on being small and fast.  Therefore jumpstart uses no shared
// libraries and no PT_INTERP.  System calls return -errno directly.
// 
// This version ignores any issues due to differences between the client
// and the server with regard to:
//	user identity (getuid, geteuid, getgid, getegid)
//	resource limitations (getrlimit)
//
// 
// THEORY OF OPERATION
//
// By the first time that ld-linux.so.2 calls _dl_debug_state() with
// _r_debug.r_state==RT_CONSISTENT, then rtld has setup the shared libraries.
// You can see this for yourself with gdb:
//	$ gdb /bin/date
//	(gdb) set stop-on-solib-events 1
//	(gdb) run
//	Stopped due to shared library event   ## RT_ADD
//	(gdb) continue
//	Stopped due to shared library event   ## RT_CONSISTENT
//	(gdb) info shared
//	From        To          Syms Read   Shared Object Library
//	0x00512770  0x005282bf  Yes         /lib/ld-linux.so.2
//	0x00a2b060  0x00a3051c  Yes         /lib/librt.so.1
//	0x00544b80  0x00632dd8  Yes         /lib/libc.so.6
//	0x0099b290  0x009a31e8  Yes         /lib/libpthread.so.0
//	(gdb)  ## just before running the app's initializers and main()
//
// However, relocation processing has not been done, which is a bug.
// See  http://sourceware.org/bugzilla/show_bug.cgi?id=2328 .
// Also, some distributions of glibc compile _dl_debug_state into fewer
// than 5 bytes of code, which makes a breakpoint cumbersome.
// So, jumpstart instead modifies the entry code of glibc
//	_start:		# sysdeps/i386/dl-machine.h
//		movl %esp,%eax
//		call _dl_start
//	_dl_start_user:
// to become instead
//	_start:
//		movl %esp,%eax
//		call my_start
//	_dl_start_user:
// This code modification is very easy to do.  Elf32_Ehdr.e_entry contains
// the address of _start.  The two instructions have been that way for
// many years.  Relocation processing has been done already upon return
// from _dl_start.  So, my_start calls _dl_start, then waits for the next task.
//
// The next task is communicated through a named pipe (fifo).
// The name of the fifo is  $JUMPSTART_DIRECTORY/path , where "path"
// is the full path to the executable.  The fifo is created if it does
// not exist.  The task info is the process id and the filedescriptors
// for two pipes.  Having read the task info, then the server fork()s a
// worker child.  Using the client process id and pseudo files such as
// /proc/pid/cwd and /proc/pid/fd/J, the worker copies argc, argv,
// the environment, and auxv from the client process.
//
// A pipe is used instead of /proc/pid/cmdline, /proc/pid/env, and
// /proc/pid/auxv because pipes are more robust and more efficient than /proc.
// Accessing /proc of a different process requires a kernel rendezvous between
// two processes, while using a pipe does not.  Also, using a pipe enables
// the pointers within argv and environ to be "pre-parsed", requiring only
// a quick relocation of addresses.  In contrast, /proc does not provide any
// length information.  There is no stat() and no lseek(), so constructing
// argv[] and environ[] would be cumbersome if /proc were used.
//
// Then the client executes PTRACE_ATTACH on the child, so that the fork()ed
// server becomes a child of the client.   Unfortunately PTRACE_ATTACH
// requires synchronization using the pipes.  Otherwise the server could exit
// before the PTRACE_ATTACH succeeds, and the client would never get the
// server's exit status in order to forward it to the client's own parent
// process.  Then the variables _dl_argc, _dl_argv, _dl_auxv of ld-linux
// are re-set to new values if necessary.  This may require black magic
// because some are local symbols that are not in the dynamic symbol table.
// Then the client sends the server on its way via PTRACE_CONT.
// Meanwhile the client acts as a normal ptrace monitor, watching for
// the fork()ed server to exit.
//
// One consequence of using ptrace is that the server cannot be ptraced by
// another process; gdb fails.  Trying to use strace on jumpstart does not
// work, either.
//
//
// FLOW DIAGRAM  (ASCII art: view or print using a fixed-width font):
//
//  $ jumpstart /bin/rm foo  ## first jumpstart for "/bin/rm"
//             |
//      create fifo $JUMPSTART_DIRECTORY/bin/rm
//             |
//        first_fork()
//            /\    #
//           /  \    #
//      parent  child
//       |        \                            [jumpstart server for /bin/rm]
//       |         +------------------------------------------+
//       |                                                    |
//       |                                             do_xmap("/bin/rm")
//       |                                                    |
//       |                                      do_xmap("/lib/ld-linux.so.2")
//       |                                                    |
//       |                                         hook "call _dl_start"
//       |                                                    |
//       |                                            goto rtld entry
//       |                                                    |
//   act like new client                             my_start(void *esp)
//       |                                                    |
//  $ jumpstart /bin/rm bar  ## new client                    |
//       |                                                    |
//  write {client_pid, pipe fds (J, K)}-->>--->>--+-->>----read fifo
//    to fifo $JUMPSTART_DIRECTORY/bin/rm         |           |
//       |                                        ^           |
//  write {argc,argv,env}-->>--+                  ^         fork()
//    to pipe fd J              \                 |          /\    #
//       |                       \                |         /  \    #
//       |                        \               |    parent  child
//       |                         \              |       /      |
//       |                          \             +--<<--+       |
//       |                           \                           |
//       |                            \              cd /proc/client_pid/cwd
//       |                             \                         |
//  read server_pid from pipe fd K <<--- -------------<< write server_pid
//       |                               \           to /proc/client_pid/fd/K
//       |                                \                      |
//       |                                 +---->>-----read {argc,argv,env}
//       |                                         from /proc/client_pid/fd/J
//       |                                                       |
//       |                                          setup argc,argv,env,auxv
//   PTRACE_ATTACH server_pid                                    |
//       |                                                       |
//   write SYNC to fd J >>-------------->> read SYNC from /proc/client_pid/fd/J
//       |                                                       |
//   server is now a child of client!                            |
//       |                                                       |
//   loop {monitor ptrace}                                  continue from
//       |                                                    my_start
//       |                                                       |
//       +---<<------------------------------------------<<---sys_exit
//       |                                                       |
//   sys_exit(server_status)                                   {poof}
//
// END OF FLOW DIAGRAM
//

typedef unsigned char uchar;

// Modifying code depends on  the machine architecture.
// This implementation is for x86.

#define EAX 0
#define ECX 1
#define EDX 2
#define EBX 3
#define ESP 4
#define EBP 5
#define ESI 6
#define EDI 7

#define rAL 0
#define rCL 1
#define rDL 2
#define rBL 3
#define rAH 4
#define rCH 5
#define rDH 6
#define rBH 7

#  define ADD 0
#  define OR  1
#  define ADC 2
#  define SBB 3
#  define AND 4
#  define SUB 5
#  define XOR 6
#  define CMP 7

#  define B2M 0
#  define W2M 1
#  define B2R 2
#  define W2R 3
#  define B2A 4
#  define W2A 5

#define CALL  0xE8
#define CALLI 2
#define FETCH 0x8B
#define FETCHB 0x8A
#define IMMB  0x80
#define IMML  0x81
#define IMMSBL 0x83
#define INT3  0xCC
#define IRET  0xCF
#define JMP8  0xEB
#define JMP32 0xE9
#define JMPI  4
#define JXX   0x70
#define LDA   0xA1
#define LEA   0x8D
#define NOP   0x90
#define POP   0x8F
#define POPA  0x61
#define POPR  0x58
#define PUSHA  0x60
#define PUSHI  0x68
#define PUSHI8 0x6A
#define PUSHM  6
#define PUSHR  0x50
#define RET   0xC3
#define RETP  0xC2
#define SETXX 0x900F
#define STA   0xA3
#define STORE 0x89
#define STOREB 0x88
#define TESTB 0x84
#define TESTW 0x85
#define XCHGM 0x87
#define XCHGA 0x90

#define JO  0x0
#define JNO 0x1
#define JC  0x2
#define JNC 0x3
#define JZ  0x4
#define JNZ 0x5
#define JBE 0x6
#define JA  0x7
#define JS  0x8
#define JNS 0x9
#define JP  0xA
#define JNP 0xB
#define JL  0xC
#define JGE 0xD
#define JLE 0xE
#define JG  0xF
#define JE JZ
#define JNE JNZ

#define D0 0
#define D8 1
#define D32 2
#define RR 3
#define A32 0  /* EBP special case */

#define S1 0
#define S2 1
#define S4 2
#define S8 3

#define MAB(m,a,b) (((m)<<6) | ((a)<<3) | (b))
#define SIB(s,i,b) (((s)<<6) | ((i)<<3) | (b))

#ifndef DEBUG  /*{*/
#define DEBUG 0
#endif  /*}*/

#include <string.h>
extern void *my_memset(void *, int, unsigned);

static 
char const *
my_strchr(char const *p, int const c)
{
	for (;; ++p) {
		if (c==*p) {
			return p;
		}
		if (0==*p) {
			return 0;
		}
	}
	return 0;
}

static
int
strnneq(char const *ap, char const *bp, int n)
{
	for (; n; --n) {
		int const a = *ap++;
		int const b = *bp++;
		int const d = b - a;
		if (0!=d || 0==a || 0==b) {
			return d;
		}
	}
	return 0;
}

static
int
strneq(char const *ap, char const *bp)
{
	for (;;) {
		int const a = *ap++;
		int const b = *bp++;
		int const d = b - a;
		if (0!=d || 0==a || 0==b) {
			return d;
		}
	}
	return 0;
}

static
char const *
getenv(char const *const name, char **__environ)
{
	unsigned const len = strlen(name);
	char **envp = __environ;
	for (; 0!=*envp; ++envp) {
		char const *const envstr = *envp;
		if (0==strnneq(envstr, name, len)
		&&  '='==envstr[len] ) {
			return 1+ len + envstr;
		}
	}
	return 0;
}

#include <alloca.h>
#include <elf.h>
#include <link.h>
#include <fcntl.h>
#include <unistd.h>
#include <asm/errno.h>
#include <sys/stat.h>
#ifndef ERESTARTSYS  /*{*/
#define ERESTARTSYS 0x200
#endif  /*}*/

extern int my_mknod(char const *path, unsigned mode, unsigned dev);

static int __attribute__((noreturn))  // forward
die(unsigned neg_errno, char const *msg);

static char *
unsimal(char *buf, unsigned u)
{
	if (10<=u) {
		buf = unsimal(buf, u/10);
	}
	*buf++ = '0' + (u % 10);
	return buf;
}

static char *
heximal(char *buf, unsigned u)
{
	int j;
	for (j=32; (j-=4)>=0; ) {
		*buf++ = "0123456789abcdef"[0xf & (u>>j)];
	}
	return buf;
}

static char *
decimal(char *buf, int w)
{
	if (w < 0) {
		*buf++ = '-';
		w = -w;
	}
	return unsimal(buf, w);
}

static int
sprintf(char *buf, char const *fmt, ...)
{
	char *const buf0 = buf;
	unsigned *argk = 1 + (unsigned *)&fmt;
	char c;
	while (0!=(c= *fmt++)) {
		if ('%'!=c) {
			*buf++ = c;
			continue;
		}
		switch (c= *fmt++) {
		default: {
			die(-127, -2+ fmt);
		} break;
		case '%': {
			*buf++ = '%';
		} break;
		case 'd': {
			int const w = (int)*argk++;
			buf = decimal(buf, w);
		} break;
		case 'p': {
			*buf++='0'; *buf++='x';
		} // fall into case 'x'
		case 'x': {
			unsigned const w = (unsigned)*argk++;
			buf = heximal(buf, w);
		} break;
		case 's': {
			char const *p = (char const *)*argk++;
			while ((*buf++ = *p++));
			--buf;
		} break;
		case 'u': {
			unsigned const w = (unsigned)*argk++;
			buf = unsimal(buf, w);
		} break;
		}
	}
	*buf = '\0';  // the terminating NUL

	return (buf - buf0);
}

static int __attribute__((noreturn))
die(unsigned neg_errno, char const *msg)
{
	char buf[50];
	write(2, buf, sprintf(buf, "\njumpstart[%d] Error %d: ", getpid(), - neg_errno));
	write(2, msg, strlen(msg));
	write(2, "\n", 1);
	_exit(-1);
}

extern int nanosleep(struct timespec const *, struct timespec *);

static void
delay(unsigned nsec)
{
	struct timespec nap;
	nap.tv_sec = 0;
	nap.tv_nsec = nsec;
	nanosleep(&nap, 0);
}

static void
mkpath(char *path)
{
	char *p = path;
	while ((p = (char *)my_strchr(p, '/'))) {
		*p = '\0';
		int const res = mkdir(path, 0700);
		switch (-res) {
		case 0: break;
		case EEXIST: break;
		case EPERM:
		case EFAULT:
		case EACCES:
		case ENOENT:
		case ENOTDIR:
		case ENOMEM:
		case EROFS:
		case ELOOP:
		case ENOSPC:
			die(res, path);
		}
		*p++ = '/';
	}
}

#include <sys/types.h>
#include <sys/mman.h>
#include <sys/ptrace.h>
#include <sys/wait.h>
extern int my_mprotect(void *, unsigned len, unsigned prot);  // minimum page cover

#define PAGE_SHIFT 12
#define PAGE_MASK (~0u<<PAGE_SHIFT)

#define REP8(x) \
    ((x)|((x)<<4)|((x)<<8)|((x)<<12)|((x)<<16)|((x)<<20)|((x)<<24)|((x)<<28))
#define EXP8(y) \
    ((1&(y)) ? 0xf0f0f0f0 : (2&(y)) ? 0xff00ff00 : (4&(y)) ? 0xffff0000 : 0)

// Find convex hull of PT_LOAD (the minimal interval which covers all PT_LOAD),
// and mmap that much, to be sure that a kernel using exec-shield-randomize
// won't place the first piece in a way that leaves no room for the rest.
static unsigned long  // returns relocation constant
xfind_pages(unsigned mflags, Elf32_Phdr const *phdr, int phnum)
{
    size_t lo= ~0, hi= 0, szlo= 0;
    char *addr;

    mflags += MAP_PRIVATE | MAP_ANONYMOUS;  // '+' can optimize better than '|'
    for (; --phnum>=0; ++phdr) if (PT_LOAD==phdr->p_type) {
        if (phdr->p_vaddr < lo) {
            lo = phdr->p_vaddr;
            szlo = phdr->p_filesz;
        }
        if (hi < (phdr->p_memsz + phdr->p_vaddr)) {
            hi =  phdr->p_memsz + phdr->p_vaddr;
        }
    }
    szlo += ~PAGE_MASK & lo;  // page fragment on lo edge
    lo   -= ~PAGE_MASK & lo;  // round down to page boundary 
    hi    =  PAGE_MASK & (hi - lo - PAGE_MASK -1);  // page length
    szlo  =  PAGE_MASK & (szlo    - PAGE_MASK -1);  // page length
    addr = mmap((void *)lo, hi, PROT_READ|PROT_WRITE|PROT_EXEC, mflags, 0, 0);
    //munmap(szlo + addr, hi - szlo);  // desirable if PT_LOAD non-contiguous
    return (unsigned long)addr - lo;
}

// Update Elf32_auxv
static void
auxv_up(Elf32_auxv_t *av, int const type, unsigned const value)
{
    if (av) for (;; ++av) {
	if (av->a_type==AT_IGNORE) {
		av->a_type = type;
	}  // fall through; no 'else'
        if (av->a_type==type) {
            av->a_un.a_val = value;
            return;
        }
        if (av->a_type==AT_NULL) {
            return;
        }
    }
}

static unsigned long
do_xmap(  // returns relocation constant
    int const fdi,
    Elf32_Ehdr const *const ehdr,
    Elf32_auxv_t *const av,
    Elf32_Addr *const p_entry,
    Elf32_Dyn const **const p_dynp
)
{
    Elf32_Phdr const *phdr = (Elf32_Phdr const *)(ehdr->e_phoff +
        (char const *)ehdr);
    unsigned long const reloc = xfind_pages(
        ((ET_DYN!=ehdr->e_type) ? MAP_FIXED : 0), phdr, ehdr->e_phnum);
    int j;

    *p_entry = ehdr->e_entry + reloc;
    auxv_up(av, AT_ENTRY, *p_entry);
    auxv_up(av, AT_BASE, reloc);
    auxv_up(av, AT_PHNUM, ehdr->e_phnum);
    for (j=0; j < ehdr->e_phnum; ++phdr, ++j)
    if (PT_DYNAMIC==phdr->p_type) {
        *p_dynp = (Elf32_Dyn const *)(phdr->p_vaddr + reloc);
    } else
    if (PT_LOAD==phdr->p_type && phdr->p_memsz!=0) {
        int const prot = 7 & (( (REP8(PROT_EXEC ) & EXP8(PF_X))
                               |(REP8(PROT_READ ) & EXP8(PF_R))
                               |(REP8(PROT_WRITE) & EXP8(PF_W)) )
                             >> ((phdr->p_flags & (PF_R|PF_W|PF_X))<<2) );
        size_t frag = ~PAGE_MASK & phdr->p_vaddr;  /* at lo end */
        size_t mlen = frag + phdr->p_filesz;
        char *addr = (char *)(phdr->p_vaddr - frag + reloc);
        char *const got = mmap(addr, mlen, prot,
				MAP_FIXED | MAP_PRIVATE | ((PROT_EXEC & prot) ? MAP_DENYWRITE : 0),
                fdi, phdr->p_offset - frag);

        if (got!=addr) {
fail:
            die(((unsigned)got | (unsigned)addr), "mmap");
        }
        if ((ehdr->e_phoff - phdr->p_offset) < phdr->p_filesz) {
            /* This PT_LOAD contains the real Elf32_Phdrs. */
            auxv_up(av, AT_PHDR, (unsigned)((ehdr->e_phoff - phdr->p_offset) +
                frag + got ));
        }
        if (0!=frag && phdr->p_flags & PF_W) {
            my_memset(addr, 0, frag);  /* fragment at lo end */
        }
        addr += mlen;
        frag = (-mlen) &~ PAGE_MASK;  /* distance to next page boundary */
        if (0!=frag && phdr->p_flags & PF_W) {
            my_memset(addr, 0, frag);  /* fragment at hi end */
        }
        addr += frag;
        if ((addr - got) < phdr->p_memsz) { /* need more pages, too */
            mlen = phdr->p_memsz - (addr - got);
            if (addr != mmap(addr, mlen, prot,
                    MAP_FIXED | MAP_PRIVATE | MAP_ANONYMOUS, 0, 0 ) ) {
                goto fail;
            }
        }
    }
    close(fdi);
    return reloc;
}

static
uchar const *
track_nop(uchar const *ip)
{
	for (;;) {
		unsigned const op = ip[0];
		unsigned const mrm = ip[1];
		if (NOP==op) {
			ip+=1; continue;
		}
		if ((FETCH==op || STORE==op)
		&&  MAB(RR,0,0)==(MAB(3,0,0) & mrm)
		&&  ((7&mrm)==(7&(mrm>>3))) ) { /* move reg,same_reg */
			ip+=2; continue;
		}
		if (LEA==op) {
			uchar const *t = ip;
			if (ESP==(7&mrm) && SIB(0,ESP,7&(mrm>>3))==(SIB(0,7,7)&ip[2])) {
				++t; goto disp;  /* 3,4,7-byte LEA m(,reg),same_reg */
			}
			if ((7&mrm)==(7&(mrm>>3)))  /* LEA m(reg),same_reg */
disp:
			switch (3&(mrm>>6)) {
			case 0: if (EBP!=(7&t[1])) { /* 2-byte LEA (reg),same_reg */
				ip+=2; continue;
			} break;
			case 1: if (0==t[2]) { /* 3-byte LEA 0(reg),same_reg */
				ip+=3; continue;
			} break;
			case 2: if (0==*(int *)(2+t)) { /* 6-byte LEA 0(reg),same_reg */
				ip+=6; continue;
			} break;
			}

			return ip;
		}
		if (JMP8==op && mrm<=13) { /* JMP L10; nop; ...; L10: # gcc 3.1 align */
			if ((3+mrm+ip)==track_nop(3+ip)) {
				return 3+mrm+ip;
			}
		}
		return ip;
	}
	return ip;
}

static Elf32_Shdr const *
elf_find_section_name(
	char const *const name,
	char const *const shstrtab,
	Elf32_Shdr const *shdr,
	int n_sec
)
{
	for (; 0 <=--n_sec; ++shdr) {
		if (0==strneq(name, &shstrtab[shdr->sh_name])) {
			return shdr;
		}
	}
	return 0;
}

struct Dictionary {
	unsigned addr;
	char const *name;
};

struct Dictionary i_dict[] = {
#define DL_ARGC 0
	{0, "_dl_argc"},
#define DL_ARGV 1
	{0, "_dl_argv"},
#define DL_AUXV 2
	{0, "_dl_auxv"},
#define DL_ENVP 3
	{0, "__environ"},
#define DL_LAST 4
	{0, 0}
};

static void
reloc_interp_syms(unsigned const i_reloc)
{
	struct Dictionary *p_dict = &i_dict[0];
	for (; 0!=p_dict->name; ++p_dict) {
		if (0!=p_dict->addr) {
			p_dict->addr += i_reloc;
		}
	}
	return;
}

// Only _dl_argv is a global symbol; _dl_argc, _dl_envp, _dl_auxv are local.
// Some builders and distributors of ld-linux.so.2 are kind.
// Take advantage if the linking view symbol table is in the file.
static void
lookup_interp_syms(int const fdi, Elf32_Ehdr const *const ehdr)
{
	if (0==ehdr->e_shnum || sizeof(Elf32_Shdr)!=ehdr->e_shentsize
	||  ehdr->e_shnum<=ehdr->e_shstrndx ) {
		return;
	}
	// Avoid *stat() whose ABI is ugly: "struct stat" and typedefs are hell!
	unsigned const length = lseek(fdi, 0, SEEK_END);
	void *const addr0 = mmap(0, length, PROT_READ, MAP_PRIVATE, fdi, 0);
	Elf32_Shdr const *const shdr0 = (Elf32_Shdr const *)(ehdr->e_shoff + addr0);
	Elf32_Shdr const *const sec_strndx = &shdr0[ehdr->e_shstrndx];
	char const *const shstrtab = (char const *)(addr0 + sec_strndx->sh_offset);
	Elf32_Shdr const *const secstr = elf_find_section_name(".strtab",
		shstrtab, shdr0, ehdr->e_shnum);
	Elf32_Shdr const *const secsym = elf_find_section_name(".symtab",
		shstrtab, shdr0, ehdr->e_shnum);
	if (0==secstr || 0==secsym) {
		goto done;
	}
	char const *const strtab = (char const *)(addr0 + secstr->sh_offset);

	struct Dictionary *p_dict = &i_dict[0];
	for (; 0!=p_dict->name; ++p_dict) {
		Elf32_Sym const *symp = (Elf32_Sym const *)(addr0 + secsym->sh_offset);
		int n_sym = secsym->sh_size / secsym->sh_entsize;
		for (; 0<=--n_sym; ++symp) {
			if (0==strneq(p_dict->name, &strtab[symp->st_name])) {
				p_dict->addr = symp->st_value;
				break;
			}
		}
	}
done:
	munmap(addr0, length);
}

#if 0  /*{ Unused because runtime symbol table usually omits the symbols we want. */
static unsigned
gnu_hash(char const *q)
{
	unsigned char const *p = (unsigned char const *)q;
	unsigned h;

	for (h= 5381; 0!=*p; ++p) {
		h += *p + (h << 5);
	}
	return h;
}

static
unsigned
elf_hash(char const *p)
{
	unsigned h;
	for (h= 0; 0!=*p; ++p) {
		h = *p + (h<<4);
		{
			unsigned const t = 0xf0000000u & h;
			h &= ~t;
			h ^= t>>24;
		}
	}
	return h;
}

static
Elf32_Sym const *
lookup_elf_hash(
	char const *const name,
	unsigned const *const htab,
	Elf32_Sym const *const dynsym,
	char const *const dynstr
)
{
	unsigned const nbucket = htab[0];
	unsigned const *const buckets = &htab[2];
	unsigned const *const chains = &buckets[nbucket];
	unsigned const m = elf_hash(name) % nbucket;
	unsigned si;
	for (si= buckets[m]; 0!=si; si= chains[si]) {
		char const *const p= dynsym[si].st_name + dynstr;
		if (0==strneq(name, p)) {
			return &dynsym[si];
		}
	}
	return 0;
}

static
Elf32_Sym const *
lookup_gnu_hash(
	char const *const name,
	unsigned const *const htab,
	Elf32_Sym const *dynsym,
	char const *const dynstr
)
{
	unsigned const n_bucket = htab[0];
	unsigned const symbias  = htab[1];
	unsigned const n_bitmask = htab[2];
	unsigned const gnu_shift = htab[3];
	unsigned const *const bitmask = &htab[4];
	unsigned const *const buckets = &bitmask[n_bitmask];

	unsigned const h = gnu_hash(name);
	unsigned const hbit1 = 037& h;
	unsigned const hbit2 = 037& (h>>gnu_shift);
	unsigned const w = bitmask[(n_bitmask -1) & (h>>5)];

	if (1& (w>>hbit1) & (w>>hbit2)) {
		unsigned bucket = buckets[h % n_bucket];
		if (0!=bucket) {
			unsigned const *const hasharr = &buckets[n_bucket];
			unsigned const *hp = &hasharr[bucket - symbias];

			dynsym += bucket;
			do if (0==((h ^ *hp)>>1)) {
				char const *const p = dynsym->st_name + dynstr;
				if (0==strneq(name, p)) {
					return dynsym;
				}
			} while (++dynsym, 0==(1u& *hp++));
		}
	}
	
	return 0;
}
#endif  /*}*/

static int
ufork(char const *const msg)
{
	int n_retry = 0;
	for (;;) {
		int const pid = fork();
		switch (-pid) {
		case EAGAIN:
		case ENOMEM: {
			if (100<=n_retry++) {
				die(pid, "fork");
			}
			delay(40*1000*1000);  // 0.04 sec
			continue;
		}
		}
#if 0!=DEBUG  /*{*/
		if (0!=pid) {
			char buf[100];
			sprintf(buf, "parent %d  child %d  %s\n", getpid(), pid, msg);
			write(2, buf, strlen(buf));
		}
#endif  /*}*/
		return pid;
	}
}

static int
uopen(char const *path, int flags, int mode)
{
	int const fd = open(path, flags, mode);
	if (fd < 0) {
		die(fd, path);
	}
	return fd;
}

static void
upipe(int fds[2])
{
	for (;;) {
		int const res = pipe(fds);
		if (-ENFILE==res) { // system file table temporarily full
			delay(40*1000*1000);  // 0.04 sec
			continue;
		}
		if (res < 0) {
			die(res, "pipe");
		}
		break;
	}
	return;
}

static unsigned
write_pipe(int const fd, void const *buf, unsigned len)
{
#if 0!=DEBUG  /*{*/
	char dbg[80];
	sprintf(dbg, "\nwrite_pipe %d %p %u\n", fd, buf, len);
	write(2, dbg, strlen(dbg));
#endif  /*}*/
	unsigned w = 0;
	while (0!=len) {
		unsigned res = write(fd, buf, len);
		if (0xfffff000<=res) switch (-res) {
		case EBADF:
		case EINVAL:
		case EFAULT:
		case EFBIG:
		case EPIPE:
		case EAGAIN:
		case EINTR:
		case ENOSPC:
		case EIO:
		default:
			die(res, "write");
		}
		w   += res;
		len -= res;
		buf += res;
	}
	return w;
}

static unsigned
read_pipe(int const fd, void *buf, unsigned len, char const *msg)
{
#if 0!=DEBUG  /*{*/
	char dbg[80];
	sprintf(dbg, "\nread_pipe %d %p %u\n", fd, buf, len);
	write(2, dbg, strlen(dbg));
#endif  /*}*/
	unsigned w = 0;
	while (0!=len) {
		unsigned res = read(fd, buf, len);
		if (0xfffff000<=res || 0==res) switch (-res) {
		case EINTR: { // ASSUME it is atomic!
			continue;
		} break;
		case EBADF:
		case EINVAL:
		case EFAULT:
		case EFBIG:
		case EPIPE:
		case EAGAIN:
		case ENOSPC:
		case EIO:
		case 0:
		default:
			die(res, msg);
		}
		w   += res;
		len -= res;
		buf += res;
	}
	return w;
}

struct Arg_data {
	void **a_argc;  // at client
	char *argv0;  // at client
	unsigned pid_client;
	unsigned char fm_client;  // pipe fd for reading
	unsigned char to_client;  // pipe fd for writing
	unsigned char len_platform;  // strlen(AT_PLATFORM)
	unsigned char dummy;
	unsigned n_byte;  // in platform,argv,envp strings
	unsigned n_argv;
	unsigned n_envp;
	unsigned n_auxv;
	unsigned long long fd_pos[3];  // lseek64 position of stdin, stdout, stderr
	unsigned long fcntl_bits[3];  // F_GETFL of stdin, stdout, stderr
};

struct r_debug *r_debug;
void **a_argc;  // set by tail_merge
static Elf32_Addr i_entry;
Elf32_Addr i_va_data;
unsigned i_sz_data;
static void *i_relro;
static unsigned i_lenro;
static int fm_interp;  // pipe fd from find_i_p_argc
static int cmd_fifo;

static void dump(void const *q, int n, char const *msg)
{
#if 0!=DEBUG  /*{*/
	unsigned const *p = (unsigned const *)q;
	write(2, msg, strlen(msg));
	while (0 < n) {
		char buf[100];
		char *q = buf;
		q += sprintf(q, "%p:", p);
		int j;
		for (j=8; --j>=0; n-=sizeof(*p), ++p) {
			q += sprintf(q, " %x", *p);
		}
		*q++= '\n'; *q = 0;
		write(2, buf, q - buf);
	}
#endif  /*}*/
}

extern void reloc_vecs(void **a_argc, unsigned reloc) __attribute__((regparm(2)));

#define SIGCHLD		17

typedef void (*__sighandler_t)(int);

struct my_sigaction {
	void (*my_handler)(int);
	unsigned long sa_flags;
	void (*sa_restorer)(void);
	unsigned sa_mask[2];		/* mask last for extensibility */
};

extern unsigned rt_sigaction(
	int signum,
	struct my_sigaction const *new,
	struct my_sigaction       *old,
	unsigned sz_sigset
);

void
handle_SIGCHLD(int signum)
{
	int status;
	/*unsigned const child =*/ waitpid(-1, &status, WNOHANG);
	return;
}

/* This matches struct stat64 in glibc2.1, hence the absolutely
 * insane amounts of padding around dev_t's.
 */
struct my_stat64 {
	unsigned short	myst_dev;
	unsigned char	__pad0[10];

#define STAT64_HAS_BROKEN_ST_INO	1
	unsigned long	__myst_ino;

	unsigned int	myst_mode;
	unsigned int	myst_nlink;

	unsigned long	myst_uid;
	unsigned long	myst_gid;

	unsigned short	myst_rdev;
	unsigned char	__pad3[10];

	long long	myst_size;
	unsigned long	myst_blksize;

	unsigned long	myst_blocks;	/* Number 512-byte blocks allocated. */
	unsigned long	__pad4;		/* future possible myst_blocks high bits */

	unsigned long	myst_atime;
	unsigned long	__pad5;

	unsigned long	myst_mtime;
	unsigned long	__pad6;

	unsigned long	myst_ctime;
	unsigned long	__pad7;		/* will be high 32 bits of ctime someday */

	unsigned long long	myst_ino;
};

extern int  my_stat64(char const *name, struct my_stat64 *);
extern int my_lstat64(char const *name, struct my_stat64 *);
extern int lseek64(int fd, unsigned hi, unsigned lo, unsigned long long *result, unsigned whence);

static Elf32_Addr
i_find_key(void *key, void const **d_interp, void const **const d_end)
{
	Elf32_Addr res = 0;
	int n = 0;
	for (; d_interp < d_end; ++d_interp) if (key==(void *)*d_interp) {
		res = (Elf32_Addr)d_interp;
		++n;
	}
	if (1==n) return res;
	return 0;
}

static void
find_rtld_dl_symbols(void const **esp)
{
	int const argc = (int)*esp;
	void const **q = esp;
	void const **const argv = 1+ esp;  while (0!=*q++) ;
	void const **const envp = q;       while (0!=*q++) ;
	void const **const auxv = q;

	void const **const p     = (void const **)i_va_data;
	void const **const p_end = (void const **)(i_sz_data + i_va_data);
	// _dl_argv, _dl_envp, _dl_auxv are pointers into the stack region.
	// The values match argv, envp, auxv.
	// We assume a unique match for each.  Find it.
	i_dict[DL_ARGV].addr = i_find_key(argv, p, p_end);
	i_dict[DL_ENVP].addr = i_find_key(envp, p, p_end);
	i_dict[DL_AUXV].addr = i_find_key(auxv, p, p_end);

	// _dl_argc is typically a small positive integer.  Many locations
	// in the .data+.bss of ld-linux.so.2 can alias that value.  Find the
	// correct location by differential analysis.  find_i_p_argc() has
	// fork()ed a scratch _dl_start() with (1+ argc), searched the .data+.bss
	// for (1+ argc), and written all matching locations to pipe fm_interp.
	// The location which matches (argc) in the current ld-linux.so.2
	// must be &_dl_argc.
	Elf32_Addr n_addr;
	read_pipe(fm_interp, &n_addr, sizeof(n_addr), "interp addr");
	unsigned n_byte = sizeof(*p) * n_addr;
	int **r = (int **)alloca(n_byte);
	read_pipe(fm_interp, r, n_byte, "interp addr");
	close(fm_interp);

	// FIXME 2007-09-02: could more than one location hold (argc)?
	for (; --n_addr>=0; ++r) if (argc==**r) {
		i_dict[DL_ARGC].addr = (Elf32_Addr)*r;
		break;
	}
}

void *(__attribute__((regparm(1))) *dl_start)(void *esp);  // returns user entry

static void * __attribute__((regparm(1)))
my_start(void const **esp)
{
#if 0!=DEBUG  /*{*/
	{char buf[80]; sprintf(buf, "my_start[%d] A\n", getpid()); write(2, buf, strlen(buf));}
#endif  /*}*/

	void *const u_entry = dl_start(esp);  // do the displaced CALL

	struct my_sigaction sa_new;
	my_memset(&sa_new, 0, sizeof(sa_new));
	sa_new.my_handler = handle_SIGCHLD;
	rt_sigaction(SIGCHLD, &sa_new, 0, 8);

	// If we did not find the address of _dl_argc in any symbol table,
	// then we must find the variables heuristically.
	if (0==i_dict[DL_ARGC].addr) {
		find_rtld_dl_symbols(esp);
	}
#if 0!=DEBUG  /*{*/
	{char buf[80]; sprintf(buf, "my_start[%d] B\n", getpid()); write(2, buf, strlen(buf));}
#endif  /*}*/

	struct Arg_data arg_data;
	for (;;) {
		read_pipe(cmd_fifo, &arg_data, sizeof(arg_data), "cmd_fifo");
		dump(&arg_data, sizeof(arg_data), "\narg_data\n");
		if (0==ufork("forkA")) {
			close(cmd_fifo);
			rt_sigaction(SIGCHLD, 0/*SIG_DFL*/, 0, 8);
			break;  // child will process arg_data
		}
		// parent waits for next request
	}
	char buf[40];
	sprintf(buf, "/proc/%d/", arg_data.pid_client);
	char *const p = &buf[strlen(buf)];

	strcpy(p, "cwd");
	int const cwd = uopen(buf, O_RDONLY, 0);
	fchdir(cwd);
	close(cwd);

	// Clone client fd 0,1,2
	unsigned fd;
	for (fd = 0; fd<=2; ++fd) {
		struct my_stat64 stat64_buf;
		sprintf(p, "fd/%d", fd);
		int res = my_lstat64(buf, &stat64_buf);
		close(fd);
		if (0==res) { // client fd is open
			int const nfd = open(buf, arg_data.fcntl_bits[fd], 0);
			if (nfd < 0) {
				die(nfd, buf);
			}
			if (fd!=nfd) {
				dup2(nfd, fd);
				close(nfd);
			}
			if (arg_data.fd_pos[fd]<=-0x1000ull) { // client could lseek64
				unsigned long long pos;
				int const err = lseek64(fd,
				(unsigned)(arg_data.fd_pos[fd]>>32),
				(unsigned)(arg_data.fd_pos[fd]     ), &pos, SEEK_SET);
				if (0!=err) {
					die(err, "lseek");
				}
			}
		}
	}

	// Open fm_client "early" so that client can close its end.
	sprintf(p, "fd/%d", arg_data.fm_client);
	int const fm_client = uopen(buf, O_RDONLY, 0);

	sprintf(p, "fd/%d", arg_data.to_client);
	int const to_client = uopen(buf, O_WRONLY, 0);
	int my_pid = getpid();
	write_pipe(to_client, &my_pid, sizeof(my_pid));
	// Client can close its ends after read() of my_pid.

	close(to_client);

	int const sz_vec = sizeof(char *) * (1+ 
		arg_data.n_argv + 1 +
		arg_data.n_envp + 1 + 
		arg_data.n_auxv * 2 + 2);
	read_pipe(fm_client, a_argc, arg_data.n_byte + sz_vec, "argv,envp");
	dump(a_argc, arg_data.n_byte + sz_vec, "\nraw data\n");

	reloc_vecs(1+ a_argc, sz_vec + (unsigned)a_argc - (unsigned)arg_data.argv0);
	dump(a_argc, arg_data.n_byte + sz_vec, "\nrelocated\n");

	// Update private variables of PT_INTERP ld-linux.so.2.
	if (0!=i_relro) {
		my_mprotect(i_relro, i_lenro, PROT_READ|PROT_WRITE);
	}
	if (0!=i_dict[DL_ARGC].addr) {
		  *(int *)i_dict[DL_ARGC].addr = *(int *)esp;
	}
	if (0!=i_dict[DL_ARGV].addr) {
		*(char ***)i_dict[DL_ARGV].addr = (char **)(1+ esp);
	}
	if (0!=i_dict[DL_ENVP].addr) {
		char **q =
		*(char ***)i_dict[DL_ENVP].addr = (char **)(1+ *(int *)esp + 1+ esp);
		if (0!=i_dict[DL_AUXV].addr) {
			while (*q++) ;
			*(char ***)i_dict[DL_AUXV].addr = q;
		}
	}
	if (0!=i_relro) {
		my_mprotect(i_relro, i_lenro, PROT_READ           );
	}

	// Don't proceed until we have been PTRACE_ATTACH
	// and thus become the child of the client.
	unsigned status = 0;
	read_pipe(fm_client, &status, sizeof(status), "attach worker");
	close(fm_client);

	return u_entry;  // result of displaced CALL
}

extern void scan_a_argc(void **esp);
extern void __attribute__((regparm(3))) jump_exit(void **esp, int, void *target);

static void __attribute__((regparm(1)))
find_i_p_argc(void **esp) 
{
	int fd[2];
	upipe(fd);
	if (0!=ufork("forkB")) { // parent
		close(fd[1]);
		fm_interp = fd[0];
		jump_exit(esp, 0, (void *)i_entry);  // leads to my_start(esp)
			// and continutation at _dl_start_user
	}

	// child
	int const new_argc = 1+ (int)*esp;  // 1+ argc
	esp[-1] = (void *)new_argc;
	dl_start(-1+ esp);  // one extra argument

	// Search .data+.bss for new_argc.  Keep linear list of matching locations.
	int const **const found = (int const **)alloca(i_sz_data);
	int const *p = (int const *)i_va_data;
	int const *const p_end = (int const *)(i_sz_data + i_va_data);
	int n_found = 1;
	for (; p < p_end; ++p) {
		if (new_argc==*p) {
			found[n_found++] = p;
		}
	}
	// Write list of matching locations to fd[1].
	found[0] = (int const *)(-1+ n_found);
	write_pipe(fd[1], found, sizeof(*found) * n_found);
#if 0!=DEBUG  /*{*/
	{char buf[80]; sprintf(buf, "find_i_p_argc %d\n", n_found); write(2, buf, strlen(buf));}
#endif  /*}*/
	_exit(0);
}

extern void isync(void *, unsigned len);  // icache sync from dcache

static Elf32_Addr
first_fork(char const *const argv0, Elf32_auxv_t *const av)
{
	int pid = ufork("forkC");
	if (0!=pid) {
		return 0;  // parent queues on fifo
	}

	// child
	int const fd = uopen(argv0, O_RDONLY, 0);
	unsigned char buf[512];
	int const res1 = read_pipe(fd, buf, sizeof(buf), "Elf_Ehdr");
	if (sizeof(buf)!=res1) {
		die(res1, argv0);
	}
	Elf32_Ehdr const *const ehdr = (Elf32_Ehdr const *)buf;
	Elf32_Addr entry;
	Elf32_Dyn const *dynp;
	unsigned const reloc = do_xmap(fd, ehdr, av, &entry, &dynp);
	Elf32_Phdr const *phdr = (Elf32_Phdr const *)(ehdr->e_phoff + (char const *)ehdr);
	unsigned const phnum = ehdr->e_phnum;
	int j;
	for (j=0; j < phnum; ++phdr, ++j) if (PT_INTERP==phdr->p_type) {
		char const *fn_interp = reloc + (char const *)phdr->p_vaddr;
		int const fdi = uopen(fn_interp, O_RDONLY, 0);
		int const res2 = read_pipe(fdi, buf, sizeof(buf), "Elf_Ehdr");
		if (sizeof(buf)!=res2) {
			die(res2, fn_interp);
		}
		lookup_interp_syms(fdi, ehdr);
		unsigned const i_reloc = do_xmap(fdi, ehdr, 0, &entry, &dynp);
		auxv_up(av, AT_BASE, i_reloc);
		reloc_interp_syms(i_reloc);

		phdr = (Elf32_Phdr const *)(ehdr->e_phoff + (char const *)ehdr);
		for (j=0; j < phnum; ++phdr, ++j) switch (phdr->p_type) {
		case PT_LOAD: if (0==i_va_data && 0!=(PF_W & phdr->p_flags)) {
			i_va_data = i_reloc + phdr->p_vaddr;
			i_sz_data = phdr->p_memsz;
		} break;
		case PT_GNU_RELRO: {
			i_relro = (void *)(i_reloc + phdr->p_vaddr);
			i_lenro = phdr->p_memsz;
		} break;
		}
		break;
	}
	unsigned char *p = (void *)entry;
	for (j=10; --j>=0; ++p) if (CALL==*p) {
		my_mprotect(p, 5, PROT_READ|PROT_WRITE);
		p+=5;
		dl_start = (void *)(((unsigned *)p)[-1] + (unsigned)p);
		((unsigned *)p)[-1] = (unsigned)my_start - (unsigned)p;
		p-=5;
		my_mprotect(p, 5, PROT_READ|PROT_EXEC);
		break;
	}
	if (j<=0) {
		die(-127, "_dl_start");
	}
	return entry;
}

static int
udup(int fd)
{
	int const nfd = dup(fd);
	if (nfd < 0) {
		die(nfd, "dup");
	}
	return nfd;
}

static unsigned
avoid_fd012(unsigned bad_bits, int *pfd, int nfd)
{
	for (; --nfd>=0; ++pfd) {
		int fd = *pfd;
		while (2>=fd) {
			bad_bits |= 1<<fd;
			fd = udup(fd);
		}
		*pfd = fd;
	}
	return bad_bits;
}

extern int
__attribute__((regparm(3)))
tail_merge(char **argv, void *continuation, char const *c_end);

int
__attribute__((regparm(1)))
my_main(char **const esp)
{
	struct Arg_data arg_data;
	int const argc = -1+ (int)esp[0];
	esp[1] = (char *)argc;
	char **const argv = 2+ esp;
	char **const __environ = &argv[1+ argc];
	arg_data.pid_client = getpid();
	arg_data.n_argv = argc;

	char const **const envpp0 = (char const **)&argv[1+ argc];
	char const **      envpp  = envpp0;
	while (0!=*envpp) ++envpp;
	arg_data.n_envp = envpp - envpp0;

	char const *c_end = 1+ strlen(envpp[-1]) + envpp[-1];
	c_end += 1+ strlen(c_end);
	arg_data.n_byte = c_end - esp[1+ 1];
	arg_data.a_argc = (void **)(1+ esp);
	arg_data.argv0 = argv[0];

	Elf32_auxv_t *const auxvp0 = (Elf32_auxv_t *)(1+ envpp);
	Elf32_auxv_t *      auxvp  = auxvp0;
	unsigned platform = 0;
	arg_data.len_platform = 0;
	for (; AT_NULL!=auxvp->a_type; ++auxvp) {
		if (AT_PLATFORM==auxvp->a_type) {
			platform = auxvp->a_un.a_val;
			// no break, we're counting all of them
		}
	}
	if (0!=platform) {
		arg_data.len_platform = strlen((char const *)platform);
		unsigned nc = 1+ arg_data.len_platform;
		unsigned d = (~3& (nc + platform)) - (~3& platform);
		auxvp += (d + sizeof(Elf32_auxv_t) -1) / sizeof(Elf32_auxv_t);
	}
	arg_data.n_auxv = auxvp - auxvp0;

	char const *const js_dirname = getenv("JUMPSTART_DIRECTORY", __environ);
	if (0==js_dirname) {
		die(-ENOENT, "JUMPSTART_DIRECTORY");
	}

	int fd;
	for (fd=0; fd<=2; ++fd) { // get current position of stdin, stdout, stderr
		int res;
		res = lseek64(fd, 0, 0, &arg_data.fd_pos[fd], SEEK_CUR);
		if (res < 0) {
			arg_data.fd_pos[fd] = (long long)res;  // sign extend
		}
		res = fcntl(fd, F_GETFL);
		arg_data.fcntl_bits[fd] = (0<=res) ? res : 0;
	}
	int const cwd = uopen(".", O_RDONLY, 0);
	int const res1 = chdir(js_dirname);
	if (res1 < 0) {
		die(res1, js_dirname);
	}
	char *const path = ('/'==*argv[0]) + argv[0];
	int own_fifo = 0;
	for (;;) {
		cmd_fifo = open(path, O_RDWR, 0);
		if (0<=cmd_fifo) { // success
			break;
		}
		switch (-cmd_fifo) {
		default:
			die(cmd_fifo, argv[0]);
		case EPERM:
		case ENOENT: {
			int const res2 = my_mknod(path, 0600 | S_IFIFO, 0);
			switch (-res2) {
			case 0: { // mknod succeeded
				own_fifo = 1;
			} break;
			case EEXIST: { // race: somebody else won
				// empty
			} break;
			case ENOENT: {
				mkpath(path);
			} break;
			case EPERM:
			case EINVAL:
			case EFAULT:
			case EACCES:
			case ENOTDIR:
			case ENOMEM:
			case EROFS:
			case ELOOP:
			case ENOSPC:
			default:
				die(res2, argv[0]);
			}
		} break;
		}
	}
	fchdir(cwd); close(cwd);

	if (0!=own_fifo) {
		i_entry = first_fork(argv[0], auxvp0);
		if (0!=i_entry) { // child
			return tail_merge(-1+ argv, 
				((0!=i_dict[DL_ARGC].addr)
					? (void *)i_entry
					: find_i_p_argc ),
				c_end );
		}
	}
	// parent

	int to_clone[2], fm_clone[2];
	upipe(to_clone);
	upipe(fm_clone);

	// Server will clone client fd 0,1,2; so we must not use them.
	{
		unsigned j;
		unsigned low_fd = 0;
		low_fd = avoid_fd012(low_fd, &to_clone[0], 2);
		low_fd = avoid_fd012(low_fd, &fm_clone[0], 2);
		low_fd = avoid_fd012(low_fd, &cmd_fifo, 1);
		for (j=0; j<=2; ++j) if (0!=((1<<j) & low_fd)) close(j);
	}

	arg_data.fm_client = to_clone[0];
	arg_data.to_client = fm_clone[1];
	write_pipe(cmd_fifo, &arg_data, sizeof(arg_data));
	close(cmd_fifo);

	int const arg_fd = to_clone[1];
	int const sz_vec = sizeof(char *) * (1+ 
		arg_data.n_argv + 1 +
		arg_data.n_envp + 1 + 
		arg_data.n_auxv * 2 + 2);
	write_pipe(arg_fd, -1+ argv, sz_vec);
	write_pipe(arg_fd, argv[0], arg_data.n_byte);

	unsigned pid_clone = 0;
	read_pipe(fm_clone[0], &pid_clone, sizeof(pid_clone), "server pid");
	close(fm_clone[0]);

	// Server has opened these, so we can close them.
	close(fm_clone[1]);
	close(to_clone[0]);

	int res3 = ptrace(PTRACE_ATTACH, pid_clone);
	if (0!=res3) {
		die(res3, "ptrace");
	}
	int status = 0;
	waitpid(pid_clone, &status, 0);

	if (WIFSTOPPED(status) && SIGSTOP==WSTOPSIG(status)) {
		// Tell server that PTRACE_ATTACH has succeeded.
		// Worker fork of server is now a child of *us*.
		write_pipe(arg_fd, &status, sizeof(status));
		close(arg_fd);
		ptrace(PTRACE_CONT, pid_clone, 0, 0);
	}
	else {
		die(-127, "attach");
	}

	for (;;) {
		waitpid(pid_clone, &status, 0);
		if (WIFEXITED(status)) { // ptrace'd child finished; so do we
			//ptrace(PTRACE_DETACH, pid_clone, 0, 0);
			_exit(WEXITSTATUS(status));
		}
		if (WIFSIGNALED(status)) { // forward to myself
			kill(arg_data.pid_client, WTERMSIG(status));
		}
		if (WIFSTOPPED(status)) { // execve(), SIGSTOP, etc.
			ptrace(PTRACE_CONT, pid_clone, 0, WSTOPSIG(status));
		}
	}
	return 0;
}

/*
vi:ts=4
*/
